package com.example.cameraxtestappjava.camera;

import android.content.Context;
import android.content.pm.PackageManager;
import android.graphics.ImageFormat;
import android.graphics.SurfaceTexture;
import android.hardware.camera2.CameraAccessException;
import android.hardware.camera2.CameraCaptureSession;
import android.hardware.camera2.CameraCharacteristics;
import android.hardware.camera2.CameraDevice;
import android.hardware.camera2.CameraManager;
import android.hardware.camera2.CameraMetadata;
import android.hardware.camera2.CaptureRequest;
import android.hardware.camera2.TotalCaptureResult;
import android.hardware.camera2.params.StreamConfigurationMap;
import android.media.Image;
import android.media.ImageReader;
import android.os.Build;
import android.os.Environment;
import android.os.Handler;
import android.os.HandlerThread;
import android.util.Log;
import android.util.Size;
import android.util.SparseIntArray;
import android.view.Surface;
import android.view.TextureView;
import android.widget.Toast;

import androidx.annotation.NonNull;
import androidx.appcompat.app.AppCompatActivity;
import androidx.core.app.ActivityCompat;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

public class SegpassCamera {
    private static final String TAG = "SegpassCamera";

    AppCompatActivity mActivity;

    private SegpassPermissionListener mPermissionListener;
    private TextureView mTextureView;
    private static final SparseIntArray ORIENTATIONS = new SparseIntArray();

    static {
        ORIENTATIONS.append(Surface.ROTATION_0, 90);
        ORIENTATIONS.append(Surface.ROTATION_90, 0);
        ORIENTATIONS.append(Surface.ROTATION_180, 270);
        ORIENTATIONS.append(Surface.ROTATION_270, 180);
    }

    private String mCameraId;
    private String mDefaultCameraId = "0";
    private Size mPreviewSize;
    private int mTotalRotation;
    protected CameraDevice mCameraDevice;
    protected String[] mCameraIds;
    protected CameraManager mCameraManager;
    protected CameraCaptureSession mCameraCaptureSessions;
    private Size mImageDimension;
    private ImageReader mImageReader;
    private File mFileFolder;
    private File mFile;
    public static final int REQUEST_CAMERA_PERMISSION = 200;
    private Handler mBackgroundHandler;
    private HandlerThread mBackgroundThread;
    private CaptureRequest.Builder mCaptureRequestBuilder;

    private static class CompareSizeByArea implements Comparator<Size> {

        @Override
        public int compare(Size ths, Size rhs) {
            return Long.signum((long) ths.getWidth() * ths.getHeight() /
                    (long) rhs.getWidth() * rhs.getHeight());
        }
    }

    public SegpassCamera(AppCompatActivity activity, TextureView textureView, SegpassPermissionListener listener) {
        mActivity = activity;
        mTextureView = textureView;
        mPermissionListener = listener;
    }

    public void init() {
        createFolder();
        setUpListeners();
    }

    private void setUpListeners() {
        mTextureView.setSurfaceTextureListener(textureListener);
    }


    public TextureView.SurfaceTextureListener textureListener = new TextureView.SurfaceTextureListener() {
        @Override
        public void onSurfaceTextureAvailable(@NonNull SurfaceTexture surface, int width, int height) {
            //open your camera here
            Log.d(TAG, "TextureView.SurfaceTextureListener.onSurfaceTextureAvailable(): the TextureView is available.");
            setupCamera(width, height);
            connectCamera();
            startPreview();
        }

        @Override
        public void onSurfaceTextureSizeChanged(@NonNull SurfaceTexture surface, int width, int height) {
            // Transform you image captured size according to the surface width and height
            Log.d(TAG, "TextureView.SurfaceTextureListener.onSurfaceTextureSizeChanged(): the TextureView size has been updated.");
            closeCamera();
            setupCamera(width, height);
            connectCamera();
            startPreview();
        }

        @Override
        public boolean onSurfaceTextureDestroyed(@NonNull SurfaceTexture surface) {
            Log.d(TAG, "TextureView.SurfaceTextureListener.onSurfaceTextureDestroyed(): the TextureView has beed destroyed.");
            return false;
        }

        @Override
        public void onSurfaceTextureUpdated(@NonNull SurfaceTexture surface) {
            // No code implementation needed for this use case.
        }
    };

    // Camera Device State Callback
    private CameraDevice.StateCallback mStateCallback = new CameraDevice.StateCallback() {
        @Override
        public void onOpened(@NonNull CameraDevice camera) {
            Log.d(TAG, "CameraDevice.StateCallback.onOpened(): camera init/open.");
            mCameraDevice = camera;
        }

        @Override
        public void onDisconnected(@NonNull CameraDevice camera) {
            Log.d(TAG, "CameraDevice.StateCallback.onDisconnected(): camera close.");
            mCameraDevice.close();
        }

        @Override
        public void onError(@NonNull CameraDevice camera, int error) {
            Log.d(TAG, "CameraDevice.StateCallback.onError(): camera close. ");
            mCameraDevice.close();
            mCameraDevice = null;
        }
    };

    //Background Thread
    public void startBackgroundThread() {
        Log.d(TAG, "startBackgroundThread()...");
        mBackgroundThread = new HandlerThread("Camera Background");
        mBackgroundThread.start();
        mBackgroundHandler = new Handler(mBackgroundThread.getLooper());
    }

    public void stopBackgroundThread() {
        Log.d(TAG, "stopBackgroundThread()...");
        mBackgroundThread.quitSafely();
        try {
            mBackgroundThread.join();
            mBackgroundThread = null;
            mBackgroundHandler = null;
        } catch (InterruptedException e) {
            Log.e(TAG, "stopBackgroundThread() - InterruptedException: " + e.getMessage());
        }
    }

    // Setup Camera
    private CameraCharacteristics getCameraCharacteristics() {
        try {
            for (String cameraId : mCameraIds) {

                // Local cc
                CameraCharacteristics cc = mCameraManager.getCameraCharacteristics(cameraId);

                // If no LENS_FACING characteristic is found, NullPointerException
                if (cc.get(CameraCharacteristics.LENS_FACING) == null)
                    throw new NullPointerException();

                if (cc.get(CameraCharacteristics.LENS_FACING) == CameraMetadata.LENS_FACING_FRONT) {
                    mCameraId = cameraId;
                } else {
                    mCameraId = mDefaultCameraId;
                }
            }
            return mCameraManager.getCameraCharacteristics(mCameraId);
        } catch (NullPointerException | CameraAccessException e) {
            Log.e(TAG, "getCameraCharacteristics() Exception: " + e.getMessage());
            return null;
        }
    }

    public void setupCamera(int width, int height) {
        mCameraManager = (CameraManager) mActivity.getSystemService(Context.CAMERA_SERVICE);

        try {
            CameraCharacteristics cameraCharacteristics = getCameraCharacteristics();
            // If no characteristics are found, throw a CameraAccessException
            if (cameraCharacteristics == null)
                throw new CameraAccessException(CameraAccessException.CAMERA_ERROR);

            // Shenanigans
            StreamConfigurationMap map = cameraCharacteristics.get(CameraCharacteristics.SCALER_STREAM_CONFIGURATION_MAP);
            int deviceOrientation = mActivity.getWindowManager().getDefaultDisplay().getRotation();
            mTotalRotation = sensorToDeviceRotation(cameraCharacteristics, deviceOrientation);
            boolean swapRotation = mTotalRotation == 90 || mTotalRotation == 270;
            int rotatedWidth = width;
            int rotatedHeight = height;
            if (swapRotation) {
                rotatedWidth = height;
                rotatedHeight = width;
            }

            assert map != null;
            mPreviewSize = chooseOptimalSize(map.getOutputSizes(SurfaceTexture.class), rotatedWidth, rotatedHeight);
        } catch (CameraAccessException e) {
            Log.e(TAG, "setupCamera() Exception: " + e.getMessage());
        }
    }

    //Connect Camera
    public void connectCameraNoListener() {
        try {
            if (ActivityCompat.checkSelfPermission(mActivity, android.Manifest.permission.CAMERA) != PackageManager.PERMISSION_GRANTED && ActivityCompat.checkSelfPermission(mActivity, android.Manifest.permission.WRITE_EXTERNAL_STORAGE) != PackageManager.PERMISSION_GRANTED) {
                ActivityCompat.requestPermissions(mActivity, new String[]{android.Manifest.permission.CAMERA, android.Manifest.permission.WRITE_EXTERNAL_STORAGE}, REQUEST_CAMERA_PERMISSION);
                return;
            }
            mCameraManager.openCamera(mCameraId, mStateCallback, mBackgroundHandler);
        } catch (CameraAccessException e) {
            Log.e(TAG, "connectCamera() CameraAccessException: " + e.getMessage());

        } catch (SecurityException e) {
            Log.e(TAG, "connectCamera() SecurityException: " + e.getMessage());
        }
    }

    public void connectCamera() {
        try {
            if (ActivityCompat.checkSelfPermission(mActivity, android.Manifest.permission.CAMERA) != PackageManager.PERMISSION_GRANTED &&
                    ActivityCompat.checkSelfPermission(mActivity, android.Manifest.permission.WRITE_EXTERNAL_STORAGE) != PackageManager.PERMISSION_GRANTED
            ) {
                mPermissionListener.onPermissionDenied();
                return;
            }
            mCameraManager.openCamera(mCameraId, mStateCallback, mBackgroundHandler);
        } catch (CameraAccessException e) {
            Log.e(TAG, "connectCamera() CameraAccessException: " + e.getMessage());

        } catch (SecurityException e) {
            Log.e(TAG, "connectCamera() SecurityException: " + e.getMessage());
        }
    }

    //  Start Preview
    public void startPreview() {
        SurfaceTexture surfaceTexture = mTextureView.getSurfaceTexture();
        assert surfaceTexture != null;
        surfaceTexture.setDefaultBufferSize(mPreviewSize.getWidth(), mPreviewSize.getHeight());
        Surface previewSurface = new Surface(surfaceTexture);

        try {
            if (mCameraDevice == null) throw new NullPointerException("Camera not started!");
            mCaptureRequestBuilder = mCameraDevice.createCaptureRequest(CameraDevice.TEMPLATE_PREVIEW);
            mCaptureRequestBuilder.addTarget(previewSurface);
            mCaptureRequestBuilder.set(CaptureRequest.CONTROL_MODE, CameraMetadata.CONTROL_MODE_AUTO);
            mCaptureRequestBuilder.set(CaptureRequest.CONTROL_VIDEO_STABILIZATION_MODE, CameraMetadata.CONTROL_VIDEO_STABILIZATION_MODE_ON);

            mCameraDevice.createCaptureSession(Collections.singletonList(previewSurface),
                    new CameraCaptureSession.StateCallback() {
                        @Override
                        public void onConfigured(@NonNull CameraCaptureSession cameraCaptureSession) {
                            try {
                                cameraCaptureSession.setRepeatingRequest(mCaptureRequestBuilder.build(), null, mBackgroundHandler);
                            } catch (CameraAccessException e) {
                                Log.e(TAG, "Camera can be accessed.");
                            }
                        }

                        @Override
                        public void onConfigureFailed(@NonNull CameraCaptureSession cameraCaptureSession) {
                            Toast.makeText(mActivity, "Configuration change", Toast.LENGTH_SHORT).show();
                        }
                    }, null);

        } catch (CameraAccessException e) {
            Log.e(TAG, "startPreview(): " + e.getMessage());
        } catch (NullPointerException | IllegalStateException e) {
            Log.e(TAG, "startPreview(): " + e.getMessage());
            resumeCamera();
        }
    }

    //Take Picture
    public void takePicture(SegpassCameraCallback cameraCallback) {
        if (mCameraDevice == null) {
            Log.e(TAG, "cameraDevice is null");
            return;
        }
        try {
            CameraCharacteristics characteristics = getCameraCharacteristics();
            if (characteristics == null)
                throw new CameraAccessException(CameraAccessException.CAMERA_ERROR);

            Size[] jpegSizes = characteristics.get(CameraCharacteristics.SCALER_STREAM_CONFIGURATION_MAP).getOutputSizes(ImageFormat.JPEG);
            if (jpegSizes == null) throw new NullPointerException("Error taking picture");

            int width = 640;
            int height = 480;

            if (jpegSizes.length > 0) {
                width = jpegSizes[0].getWidth();
                height = jpegSizes[0].getHeight();
            }

            ImageReader mReader = ImageReader.newInstance(width, height, ImageFormat.JPEG, 1);

            List<Surface> outputSurfaces = new ArrayList<Surface>(2);
            outputSurfaces.add(mReader.getSurface());
            outputSurfaces.add(new Surface(mTextureView.getSurfaceTexture()));

            final CaptureRequest.Builder captureBuilder = mCameraDevice.createCaptureRequest(CameraDevice.TEMPLATE_STILL_CAPTURE);
            captureBuilder.addTarget(mReader.getSurface());
            captureBuilder.set(CaptureRequest.CONTROL_MODE, CameraMetadata.CONTROL_MODE_AUTO);

            // Orientation
            int rotation = mActivity.getWindowManager().getDefaultDisplay().getRotation();
            captureBuilder.set(CaptureRequest.JPEG_ORIENTATION, ORIENTATIONS.get(rotation));
            String fileName = "IMG_" + System.currentTimeMillis() + ".jpg";
            mFile = new File(mFileFolder + "/" + fileName); // Ver como hacer para guardar en la carpeta de la app

            ImageReader.OnImageAvailableListener readerListener = iReader -> {
                try (Image image = iReader.acquireLatestImage()) {
                    ByteBuffer buffer = image.getPlanes()[0].getBuffer();
                    byte[] bytes = new byte[buffer.capacity()];
                    buffer.get(bytes);
                    savePicture(bytes);
                } catch (IOException e) {
                    cameraCallback.onPictureTakenFailError(e.getMessage());
                }
            };

            mReader.setOnImageAvailableListener(readerListener, mBackgroundHandler);
            final CameraCaptureSession.CaptureCallback captureListener = new CameraCaptureSession.CaptureCallback() {
                @Override
                public void onCaptureCompleted(@NonNull CameraCaptureSession session, @NonNull CaptureRequest request, @NonNull TotalCaptureResult result) {
                    super.onCaptureCompleted(session, request, result);
                    Log.d(TAG, "file");
                    //Toast.makeText(mActivity, "Saved:" + mFile, Toast.LENGTH_SHORT).show();
                    createCameraPreview();
                    cameraCallback.onPictureTakenSuccess("Saved:" + mFile);
                }
            };

            mCameraDevice.createCaptureSession(outputSurfaces, new CameraCaptureSession.StateCallback() {
                @Override
                public void onConfigured(@NonNull CameraCaptureSession session) {
                    try {
                        session.capture(captureBuilder.build(), captureListener, mBackgroundHandler);
                    } catch (CameraAccessException e) {
                        e.printStackTrace();
                    }
                }

                @Override
                public void onConfigureFailed(@NonNull CameraCaptureSession session) {
                }
            }, mBackgroundHandler);
        } catch (CameraAccessException e) {
            cameraCallback.onPictureTakenFailError(e.getMessage());
        }
    }

    private void savePicture(byte[] bytes) throws IOException {
        OutputStream output = null;
        try {
            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
                output = Files.newOutputStream(mFile.toPath());
            } else {
                output = new FileOutputStream(mFile);
            }
            output.write(bytes);
        } finally {
            if (output != null) {
                output.close();
            }
        }
    }

    protected void createCameraPreview() {
        try {
            CameraCharacteristics cc = getCameraCharacteristics();

            StreamConfigurationMap map = cc.get(CameraCharacteristics.SCALER_STREAM_CONFIGURATION_MAP);
            assert map != null;
            mImageDimension = map.getOutputSizes(SurfaceTexture.class)[0];

            SurfaceTexture texture = mTextureView.getSurfaceTexture();
            assert texture != null;
            texture.setDefaultBufferSize(mImageDimension.getWidth(), mImageDimension.getHeight());
            Surface surface = new Surface(texture);
            mCaptureRequestBuilder = mCameraDevice.createCaptureRequest(CameraDevice.TEMPLATE_PREVIEW);
            mCaptureRequestBuilder.addTarget(surface);
            mCameraDevice.createCaptureSession(Arrays.asList(surface), new CameraCaptureSession.StateCallback() {
                @Override
                public void onConfigured(@NonNull CameraCaptureSession cameraCaptureSession) {
                    //The camera is already closed
                    if (null == mCameraDevice) {
                        return;
                    }
                    // When the session is ready, we start displaying the preview.
                    mCameraCaptureSessions = cameraCaptureSession;
                    updatePreview();
                }

                @Override
                public void onConfigureFailed(@NonNull CameraCaptureSession cameraCaptureSession) {
                    Toast.makeText(mActivity, "Configuration change", Toast.LENGTH_SHORT).show();
                }
            }, null);
        } catch (CameraAccessException e) {
            e.printStackTrace();
        }
    }

    protected void updatePreview() {
        if (null == mCameraDevice) {
            Log.e(TAG, "updatePreview error, return");
        }
        mCaptureRequestBuilder.set(CaptureRequest.CONTROL_MODE, CameraMetadata.CONTROL_MODE_AUTO);
        try {
            mCameraCaptureSessions.setRepeatingRequest(mCaptureRequestBuilder.build(), null, mBackgroundHandler);
        } catch (CameraAccessException e) {
            e.printStackTrace();
        }
    }

    public void closeCamera() {
        if (mCameraDevice != null) {
            mCameraDevice.close();
            mCameraDevice = null;
        }
        if (mImageReader != null) {
            mImageReader.close();
            mImageReader = null;
        }
    }

    private static int sensorToDeviceRotation(CameraCharacteristics cameraCharacteristics, int deviceOrientation) {
        int sensorOrienatation = cameraCharacteristics.get(CameraCharacteristics.SENSOR_ORIENTATION);
        deviceOrientation = ORIENTATIONS.get(deviceOrientation);
        return (sensorOrienatation + deviceOrientation + 360) % 360;

    }

    private static Size chooseOptimalSize(Size[] choices, int textureViewWidth,
                                          int textureViewHeight) {

        // Collect the supported resolutions that are at least as big as the preview Surface
        List<Size> bigEnough = new ArrayList<>();
        // Collect the supported resolutions that are smaller than the preview Surface

        for (Size option : choices) {
            if (option.getHeight() == option.getWidth() * textureViewHeight / textureViewWidth &&
                    option.getWidth() >= textureViewWidth &&
                    option.getHeight() >= textureViewHeight) {
                bigEnough.add(option);
            }
        }
        // Pick the smallest of those big enough. If there is no one big enough, pick the
        // largest of those not big enough.
        if (!bigEnough.isEmpty()) {
            return Collections.min(bigEnough, new CompareSizeByArea());
        } else {
            Log.e(TAG, "Couldn't find any suitable preview size");
            return choices[0];
        }
    }

    private void createFolder() {
        File folder = Environment.getExternalStoragePublicDirectory(Environment.DIRECTORY_PICTURES);
        mFileFolder = new File(folder, "Camera2TestApp");
        if (!mFileFolder.exists()) {
            mFileFolder.mkdirs();
        }
    }

    public void resumeCamera() {
        startBackgroundThread();
        if (mTextureView.isAvailable()) {
            setupCamera(mTextureView.getWidth(), mTextureView.getHeight());
            connectCamera();
            startPreview();
        } else {
            mTextureView.setSurfaceTextureListener(textureListener);
        }
    }

    public void pauseCamera() {
        closeCamera();
        stopBackgroundThread();
    }

    public void destroyCamera() {
        closeCamera();
    }
}
